/* -*- c -*-
 *
 * Copyright (C) 2005-2009 Nexenta
 *
 */


#include "includes.h"
#include "librpc/rpc/dcerpc.h"
#include "librpc/gen_ndr/ndr_oxidresolver.h"
#include "librpc/gen_ndr/ndr_oxidresolver_c.h"
#include "librpc/gen_ndr/ndr_dcom.h"
#include "librpc/gen_ndr/ndr_dcom_c.h"
#include "librpc/gen_ndr/ndr_remact_c.h"
#include "librpc/gen_ndr/ndr_epmapper_c.h"
#include "librpc/gen_ndr/com_dcom.h"
#include "librpc/rpc/dcerpc_table.h"

#include "lib/com/dcom/dcom.h"
#include "lib/com/proto.h"
#include "lib/com/dcom/proto.h"

#include "wmi/wmi.h"

struct WBEMCLASS;
struct WBEMOBJECT;

#include "wmi/proto.h"

#ifdef sun
#include <sys/stat.h>
#include <sys/vnode.h>
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"


#if NET_WMI_DEBUG
#define DEBUG_MSG(...) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUG_MSG(...)
#endif


struct com_context* com_ctx = NULL;

/*
 * On 32-bit OS (and some 64-bit) Perl does not have an
 * integer type capable of storing 64 bit numbers. So
 * we serialize to/from strings on these platforms
 */

/*
wmi_int64_t
_wmi_parse_int64(SV *sv) {
#ifdef USE_64_BIT_ALL
    return SvIV(sv);
#else
    //DEBUG_MSG("Parrse %s\n", SvPV_nolen(sv));
    return strtoll(SvPV_nolen(sv), NULL, 10);
#endif
}

wmi_uint64_t
_wmi_parse_uint64(SV *sv) {
#ifdef USE_64_BIT_ALL
    return SvUV(sv);
#else
    //DEBUG_MSG("Parrse %s\n", SvPV_nolen(sv));
    return strtoull(SvPV_nolen(sv), NULL, 10);
#endif
}


#ifndef PRId64
#define PRId64 "lld"
#endif

SV *
_wmi_format_int64(wmi_int64_t val) {
#ifdef USE_64_BIT_ALL
    return newSViv(val);
#else
    char buf[100];
    int len;
    len = snprintf(buf, 100, "%" PRId64, val);
    //DEBUG_MSG("Format i64 [%" PRId64 "] to [%s]\n", val, buf);
    return newSVpv(buf, len);
#endif
}

#ifndef PRIu64
#define PRIu64 "llu"
#endif

SV *
_wmi_format_uint64(wmi_uint64_t val) {
#ifdef USE_64_BIT_ALL
    return newSVuv(val);
#else
    char buf[100];
    int len;
    len = snprintf(buf, 100, "%" PRIu64, val);
    //DEBUG_MSG("Format u64 [%" PRIu64 "] to [%s]\n", val, buf);
    return newSVpv(buf, len);
#endif
}
*/



void
_object_release(void *obj) {
    DEBUG_MSG("Releasing object count on %p\n", obj);
    SvREFCNT_dec((SV*)obj);
}

void print_hello (void)
{
    printf("hello, world\n");
}

SV*
__GetSV(union CIMVAR* obj, enum CIMTYPE_ENUMERATION cimtype)
{
	switch(cimtype)
	{
	    case CIM_SINT8: return newSViv(obj->v_sint8);
	    case CIM_UINT8: return newSVuv(obj->v_uint8);
	    case CIM_SINT16: return newSViv(obj->v_sint16);
	    case CIM_UINT16: return newSVuv(obj->v_uint16);
	    case CIM_SINT32: return newSViv(obj->v_sint32);
	    case CIM_UINT32: return newSVuv(obj->v_uint32);
	    case CIM_SINT64: return newSViv(obj->v_sint64);
	    case CIM_UINT64: return newSViv(obj->v_sint64);
	    case CIM_REAL32: return newSVnv(obj->v_real32);
	    case CIM_REAL64: return newSVnv(obj->v_real64);
	    case CIM_BOOLEAN: return newSVuv(obj->v_boolean);
	    case CIM_STRING:
	    case CIM_DATETIME:
	    case CIM_REFERENCE: return newSVpv(obj->v_string ? obj->v_string : "",0);
	    case CIM_CHAR16: return newSViv(obj->v_sint16);
	    case CIM_OBJECT: 
		{
			SV* sv;
			sv = sv_newmortal();
    			sv_setref_pv( sv, "Net::WMI::ClassObject", (void*)obj->v_object);
			return sv;
		}
	    default: return newSVsv(&PL_sv_undef);
	}
}

#define RETURN_CIM_AV(arr,func) { \
		int i; \
		SV* sv; \
		AV* av = newAV(); \
		for(i=0;i<arr->count;i++) { \
		    sv = newSV ## func(arr->item[i]); \
		    av_push(av, sv); } \
		    return av; \
		}

#define RETURN_CIM_STR_AV(arr) { \
		int i; \
		SV* sv; \
		AV* av = newAV(); \
		fprintf(stderr,"FROM CIM_STR_AV: %i\n", arr->count); \
		return av; \
		for(i=0;i<arr->count;i++) { \
		    char * str = arr->item[i] ? arr->item[i] : ""; \
		    fprintf(stderr,"FROM CIM_STR_AV: %s\n", str); \
		    sv = newSVpv(str , strlen(str)); \
		    av_push(av, sv); } \
		    return av; \
		}


AV*
__GetAV(union CIMVAR* obj, enum CIMTYPE_ENUMERATION cimtype)
{
	//fprintf(stderr,"FROM GETAV: %08X %08X\n",cimtype, obj->a_string);
	switch(cimtype)
	{
	    case CIM_ARR_SINT8: RETURN_CIM_AV(obj->a_sint8,iv);
	    case CIM_ARR_UINT8: RETURN_CIM_AV(obj->a_uint8,uv);
	    case CIM_ARR_SINT16: RETURN_CIM_AV(obj->a_sint16,iv);
	    case CIM_ARR_UINT16: RETURN_CIM_AV(obj->a_uint16,uv);
	    case CIM_ARR_SINT32: RETURN_CIM_AV(obj->a_sint32,iv);
	    case CIM_ARR_UINT32: RETURN_CIM_AV(obj->a_uint32,uv);
	    case CIM_ARR_SINT64: RETURN_CIM_AV(obj->a_sint64,iv);
	    case CIM_ARR_UINT64: RETURN_CIM_AV(obj->a_sint64,uv);
	    case CIM_ARR_REAL32: RETURN_CIM_AV(obj->a_real32,nv);
	    case CIM_ARR_REAL64: RETURN_CIM_AV(obj->a_real64,nv);
	    case CIM_ARR_BOOLEAN: RETURN_CIM_AV(obj->a_boolean,uv);
	    case CIM_ARR_STRING:
	    case CIM_ARR_DATETIME:
	    case CIM_ARR_REFERENCE: 
		{
			int i;
			SV* sv;
			AV* av = newAV();
			//fprintf(stderr, "OBJ: %08X %08X\n",obj, obj?obj->a_string:0);
			for(i=0;obj && obj->a_string && i<obj->a_string->count;i++)
			{
				//fprintf(stderr, "OBJ: %s\n", obj->a_string->item[i]);
		    		sv = newSVpv(obj->a_string->item[i] ? obj->a_string->item[i] : "" , 0);
		    		av_push(av, sv);
			}
		    	return av; \
		}
		//RETURN_CIM_STR_AV(obj->a_string);
	    case CIM_ARR_CHAR16: RETURN_CIM_AV(obj->a_sint16,iv);
	    case CIM_ARR_OBJECT:
		{
			int i;
			SV* sv;
			AV* av = newAV();
			for(i=0;obj && i<obj->a_object->count;i++)
			{
				sv = newSV(0); //sv_newmortal();
    				sv_setref_pv( sv, "Net::WMI::ClassObject", (void*)obj->a_object->item[i]);
				av_push(av, sv);
				fprintf(stderr,"FROM __GetAV: %08X\n",sv);
			}
			return av;
		}
	    default: return newAV();
	}
}


SV *
_sv_from_error (const char* error)
{
    HV *hv;

    hv = newHV ();

    /* map DBusError attributes to hash keys */
    hv_store (hv, "name", 4, newSVpv ("error", 0), 0);
    hv_store (hv, "message", 7, newSVpv (error, 0), 0);

    return sv_bless (newRV_noinc ((SV*) hv), gv_stashpv ("Net::WMI::Error", TRUE));
}

void
_croak_error (const char* error)
{
    sv_setsv (ERRSV, _sv_from_error (error));
    //croak (Nullch);
    croak(error);
}

void
_populate_constant(HV *href, char *name, int val)
{
    hv_store(href, name, strlen(name), newSViv(val), 0);
}

#define REGISTER_CONSTANT(name, key) _populate_constant(constants, #key, name)

MODULE = Net::WMI		PACKAGE = Net::WMI

PROTOTYPES: ENABLE
BOOT:
    {
	HV *constants;


	// initialization
        lp_load();
        dcerpc_init();
        dcerpc_table_init();

        dcom_proxy_IUnknown_init();
        dcom_proxy_IWbemLevel1Login_init();
        dcom_proxy_IWbemServices_init();
        dcom_proxy_IEnumWbemClassObject_init();
        dcom_proxy_IRemUnknown_init();
        dcom_proxy_IWbemFetchSmartEnum_init();
        dcom_proxy_IWbemWCOSmartEnum_init();
        dcom_proxy_IWbemClassObject_init();
        
        com_init_ctx(&com_ctx, NULL);
	dcom_client_init(com_ctx,NULL);

        DEBUG_MSG("Inited\n");
    }

void
print_hello()


MODULE = Net::WMI::Connection		PACKAGE = Net::WMI::Connection
PROTOTYPES: ENABLE


struct IWbemServices *
_ConnectServer(hostname, namespace, user, pass)
	char *hostname;
	char *namespace;
	char *user;
	char *pass;
    PREINIT:
	WERROR result;
	struct IWbemServices *con;
    CODE:	
        DEBUG_MSG("Open connection //%s %s%%%s %s\n", hostname, user, pass, namespace);
	 result = WBEM_ConnectServer(com_ctx, hostname, namespace, user, pass, 0,0,0,0,&con);
	if (!con || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	DEBUG_MSG("Open connection %p\n",con);
	RETVAL = con;
    OUTPUT:
	RETVAL

struct IEnumWbemClassObject*
_ExecQuery(service, query)
	struct IWbemServices* service;
	char *query;
    PREINIT:
	WERROR result;
	struct IEnumWbemClassObject* pEnum;
    CODE:
	DEBUG_MSG("ExecQuery: %p %s\n", service, query);
	result = IWbemServices_ExecQuery(service, (struct com_context*)service, "WQL",query,WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_ENSURE_LOCATABLE, NULL, &pEnum);
	if (!pEnum || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	RETVAL = pEnum;
    OUTPUT:
	RETVAL

struct IEnumWbemClassObject*
_ExecNotificationQuery(service, query)
	struct IWbemServices* service;
	char *query;
    PREINIT:
	WERROR result;
	struct IEnumWbemClassObject* pEnum;
    CODE:
	DEBUG_MSG("ExecQuery: %p %s\n", service, query);
	result = IWbemServices_ExecNotificationQuery(service, (struct com_context*)service, "WQL", query, WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_FORWARD_ONLY, NULL, &pEnum);
	if (!pEnum || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	RETVAL = pEnum;
    OUTPUT:
	RETVAL


	
struct IWbemClassObject*
_GetObject(service, ClassName)
	struct IWbemServices* service;
	char *ClassName;
    PREINIT:
	WERROR result;
	struct IWbemClassObject* pObj = NULL;
    CODE:
	DEBUG_MSG("GetObject: %p %s\n", service, ClassName);
	result = IWbemServices_GetObject(service, (struct com_context*)service, ClassName,WBEM_FLAG_RETURN_WBEM_COMPLETE, NULL, &pObj, NULL);
	DEBUG_MSG("GetObject: %p %i\n", pObj, result);
	if (!pObj || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	RETVAL = pObj;
    OUTPUT:
	RETVAL

struct IWbemClassObject *
_ExecMethod1(service,ObjName,MethodName)
	struct IWbemServices * service;
	char* ObjName;
	char* MethodName;
    PREINIT:
	WERROR result;
	struct IWbemClassObject* pObj = NULL;
    CODE:
	DEBUG_MSG("GetMethod: %p %s %s\n", service,ObjName,MethodName);
	result = IWbemServices_ExecMethod(service, (struct com_context*)service, ObjName, MethodName, 0, NULL, NULL, &pObj, NULL);
	if (!pObj || !W_ERROR_IS_OK(result)) {
		_croak_error (wmi_errstr(result));
	}
        RETVAL = pObj;
    OUTPUT:
	RETVAL

struct IWbemClassObject *
_ExecMethod(service,ObjName,MethodName,in)
	struct IWbemServices * service;
	char* ObjName;
	char* MethodName;
	struct IWbemClassObject* in;
    PREINIT:
	WERROR result;
	struct IWbemClassObject* pObj = NULL;
    CODE:
	DEBUG_MSG("GetMethod: %p %s %s %p\n", service,ObjName,MethodName,in);
	result = IWbemServices_ExecMethod(service, (struct com_context*)service, ObjName, MethodName, 0, NULL, in, &pObj, NULL);
	if (!pObj || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	RETVAL = pObj;
    OUTPUT:
	RETVAL

void
DESTROY(self)
    HV* self;
    PREINIT:
	SV** con;
	struct com_context* ctx = NULL;
    CODE:
	con = hv_fetch(self, "con", 3, 0);
	DEBUG_MSG("IWbemServices___:%p %p %p\n",self, con, ctx);


MODULE = Net::WMI::IWbemServices		PACKAGE = Net::WMI::IWbemServices
PROTOTYPES: ENABLE

void
DESTROY(obj)
	struct IWbemServices* obj;
    CODE:
	DEBUG_MSG("IWbemServices: %p\n", obj);
	talloc_free(obj);


MODULE = Net::WMI::EnumClassObject		PACKAGE = Net::WMI::EnumClassObject
PROTOTYPES: ENABLE

void
_Reset(pEnum)
	struct IEnumWbemClassObject* pEnum;
    PREINIT:
	WERROR result;
    CODE:
	DEBUG_MSG("Reset: %p\n", pEnum);
	result = IEnumWbemClassObject_Reset(pEnum, (struct com_context*)pEnum);
	if (!W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}

struct WbemClassObject*
_Next(pEnum)
	struct IEnumWbemClassObject* pEnum;
    PREINIT:
	WERROR result;
	struct WbemClassObject* pObj[1];
	int ret;
    CODE:
	DEBUG_MSG("Next:%p\n", pEnum);
	result = IEnumWbemClassObject_SmartNext(pEnum,(struct com_context*)pEnum,0xFFFFFFFF,1,pObj,&ret);
	if (!W_ERROR_EQUAL(result, WERR_BADFUNC)) {
	    if (!W_ERROR_IS_OK(result)) {
		 _croak_error (wmi_errstr(result));
	    }
	}
	if(ret == 1) {
	    RETVAL = pObj[0];
	}
	else
	    RETVAL = NULL;
	OUTPUT:
	    RETVAL

MODULE = Net::WMI::IEnumWbemClassObject		PACKAGE = Net::WMI::IEnumWbemClassObject
PROTOTYPES: ENABLE


MODULE = Net::WMI::ClassObject		PACKAGE = Net::WMI::ClassObject
PROTOTYPES: ENABLE


const char*
_GetClassName(obj)
	struct WbemClassObject* obj;
    PREINIT:
    CODE:
	DEBUG_MSG("GetClassName: %p \n", obj);
	RETVAL = obj->obj_class->__CLASS;
	OUTPUT:
	    RETVAL

const char*
_GetInstanceClassName(obj)
	struct WbemClassObject* obj;
    PREINIT:
	int i;
    CODE:
	DEBUG_MSG("GetClassQualifiers: %s %s\n", obj->__NAMESPACE, obj->__SERVER);
	RETVAL = obj->instance->__CLASS;
	OUTPUT:
	    RETVAL


SV*
__PATH(obj)
	struct WbemClassObject* obj;
    PREINIT:
	int i,j,k=0;
	char *val;
	TALLOC_CTX *mem_ctx;
    CODE:
	DEBUG_MSG("__PATH: \\\\%s\\%s\n", obj->__SERVER, obj->__NAMESPACE);
	RETVAL = newSV(0);
	sv_setpvf(RETVAL,"\\\\%s\\%s:%s.",obj->__SERVER, obj->__NAMESPACE,obj->obj_class->__CLASS);
	for(i=0;i<obj->obj_class->__PROPERTY_COUNT;i++)
	{
	    DEBUG_MSG("%s %i ", obj->obj_class->properties[i].name, obj->obj_class->properties[i].desc->qualifiers.count);
	    for(j=0;j<obj->obj_class->properties[i].desc->qualifiers.count;j++)
	    {
		DEBUG_MSG(" %s",obj->obj_class->properties[i].desc->qualifiers.item[j]->name);
		if(!strcmp(obj->obj_class->properties[i].desc->qualifiers.item[j]->name,"key"))
		{
		    sv_catpvf(RETVAL,"%s%s=",k?",":"",obj->obj_class->properties[i].name);
		    mem_ctx = talloc_new(0);
		    switch (obj->obj_class->properties[i].desc->cimtype & CIM_TYPEMASK)
		    {
		        case CIM_SINT8: val=talloc_asprintf(mem_ctx, "%d", obj->instance->data[i].v_sint8);break;
		        case CIM_UINT8: val=talloc_asprintf(mem_ctx, "%u", obj->instance->data[i].v_uint8);break;
		        case CIM_SINT16: val=talloc_asprintf(mem_ctx, "%d", obj->instance->data[i].v_sint16);break;
		        case CIM_UINT16: val=talloc_asprintf(mem_ctx, "%u", obj->instance->data[i].v_uint16);break;
		        case CIM_SINT32: val=talloc_asprintf(mem_ctx, "%d", obj->instance->data[i].v_sint32);break;
		        case CIM_UINT32: val=talloc_asprintf(mem_ctx, "%u", obj->instance->data[i].v_uint32);break;
		        case CIM_SINT64: val=talloc_asprintf(mem_ctx, "%lld", obj->instance->data[i].v_sint64);break;
		        case CIM_UINT64: val=talloc_asprintf(mem_ctx, "%llu", obj->instance->data[i].v_sint64);break;
		        case CIM_REAL32: val=talloc_asprintf(mem_ctx, "%f", obj->instance->data[i].v_real32);break;
		        case CIM_REAL64: val=talloc_asprintf(mem_ctx, "%f", obj->instance->data[i].v_real64);break;
		        case CIM_BOOLEAN: val=talloc_asprintf(mem_ctx, "%s", obj->instance->data[i].v_boolean?"True":"False");break;
		        case CIM_STRING:
		        case CIM_DATETIME:
		        case CIM_REFERENCE: val=talloc_asprintf(mem_ctx, "%s", obj->instance->data[i].v_string);break;
		        case CIM_CHAR16: val=talloc_asprintf(mem_ctx, "%u",obj->instance->data[i].v_sint16);break;
		        case CIM_OBJECT: val=talloc_asprintf(mem_ctx, "Unsupported2");break;
		        default:  val=talloc_asprintf(mem_ctx, "Unsupported3");
		    }
		    sv_catpvf(RETVAL,"\"%s\"",val);
		    talloc_free(mem_ctx);
		    k++;
		}
	    }
	    DEBUG_MSG("\n");
	}
	OUTPUT:
	    RETVAL


int
_GetPropertyCount(obj)
	struct WbemClassObject* obj;
    PREINIT:
    CODE:
	DEBUG_MSG("GetPropCount: %p \n", obj);
	RETVAL = obj->obj_class->__PROPERTY_COUNT;
	OUTPUT:
	    RETVAL

const char*
_GetPropertyName(obj, num)
	struct WbemClassObject* obj;
	int num;
    PREINIT:
    CODE:
	DEBUG_MSG("GetPropertyName: %p %i\n", obj, num);
	RETVAL = obj->obj_class->properties[num].name;
	OUTPUT:
	    RETVAL

int
_GetPropertyType(obj, name)
	struct WbemClassObject* obj;
	char *name;
    PREINIT:
	int i;
    CODE:
	DEBUG_MSG("GetPropertyName: %p\n", obj);
	for(i=0;i<obj->obj_class->__PROPERTY_COUNT;i++)
	{
	    if(!strcasecmp(name,obj->obj_class->properties[i].name))
	    {
		RETVAL = obj->obj_class->properties[i].desc->cimtype & CIM_TYPEMASK;
		break;
	    }
	}
	OUTPUT:
	    RETVAL


int
_GetPropertyTypeByNum(obj, num)
	struct WbemClassObject* obj;
	int num;
    PREINIT:
    CODE:
	DEBUG_MSG("GetPropertyName: %p\n", obj);
	    RETVAL = obj->obj_class->properties[num].desc->cimtype & CIM_TYPEMASK;
	OUTPUT:
	    RETVAL


union CIMVAR*
_GetProperty(obj, name)
	struct WbemClassObject* obj;
	char *name;
    PREINIT:
	int i;
    CODE:
	DEBUG_MSG("GetPropertyName: %p\n", obj);
	for(i=0;i<obj->obj_class->__PROPERTY_COUNT;i++)
	{
	    if(!strcasecmp(name,obj->obj_class->properties[i].name))
	    {
		RETVAL = &obj->instance->data[i];
		break;
	    }
	}
	OUTPUT:
	    RETVAL

union CIMVAR*
_GetPropertyByNum(obj, num)
	struct WbemClassObject* obj;
	int num;
    PREINIT:
    CODE:
	DEBUG_MSG("GetPropertyName: %p\n", obj);
	    RETVAL = &obj->instance->data[num];
	OUTPUT:
	    RETVAL


MODULE = Net::WMI::CIMVAR		PACKAGE = Net::WMI::CIMVAR
PROTOTYPES: ENABLE

SV*
_GetSV(obj,cimtype)
	union CIMVAR* obj;
	enum CIMTYPE_ENUMERATION cimtype;
    PREINIT:
    CODE:
	RETVAL = __GetSV(obj,cimtype);
    OUTPUT:
	RETVAL

AV*
_GetAV(obj,cimtype)
	union CIMVAR* obj;
	enum CIMTYPE_ENUMERATION cimtype;
    PREINIT:
    CODE:
	RETVAL = __GetAV(obj,cimtype);
    OUTPUT:
	RETVAL


MODULE = Net::WMI::IWbemClassObject		PACKAGE = Net::WMI::IWbemClassObject
PROTOTYPES: ENABLE


MODULE = Net::WMI::IClassObject		PACKAGE = Net::WMI::IClassObject
PROTOTYPES: ENABLE

void
_GetMethod(obj,method)
	struct IWbemClassObject* obj;
	char *method;
    PREINIT:
	SV* sv;
	WERROR result;
	struct IWbemClassObject* inc = NULL;
	struct IWbemClassObject* outc = NULL;
    PPCODE:
	DEBUG_MSG("In GetMethod %p %s %p %p\n",obj->object_data,method, inc,outc);
	result = IWbemClassObject_GetMethod(obj, (struct com_context*)obj, method, 0, &inc, &outc);
	if (!W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	DEBUG_MSG("In GetMethod %p %p\n",inc,outc);
	EXTEND(SP, 2);
	sv = sv_newmortal();
	sv_setref_pv( sv, "Net::WMI::IWbemClassObject", (void*)inc );
	PUSHs(sv);
	sv = sv_newmortal();
	sv_setref_pv( sv, "Net::WMI::IWbemClassObject", (void*)outc );
	PUSHs(sv);

HV*
_ListProperties(obj)
	struct IWbemClassObject* obj;
    PREINIT:
	struct WbemClassObject* d;
	int i;
	SV* sv;
    CODE:
	d = obj->object_data;
        RETVAL = newHV();
        if(d)
        {
            for(i=0;i<d->obj_class->__PROPERTY_COUNT;i++)
            {
                sv = newSVuv(d->obj_class->properties[i].desc->cimtype);
                hv_store(RETVAL,d->obj_class->properties[i].name, strlen(d->obj_class->properties[i].name), sv, 0);
                DEBUG_MSG("Pname %s %08X\n", d->obj_class->properties[i].name, d->obj_class->properties[i].desc->cimtype);
            }
        }
    OUTPUT:
	RETVAL

struct WbemClassObject*
_GetClassObject(obj)
	struct IWbemClassObject* obj;
    PREINIT:
    CODE:	
	RETVAL = obj->object_data;
    OUTPUT:
	RETVAL


struct IWbemClassObject*
_SpawnInstance(obj)
	struct IWbemClassObject* obj;
    PREINIT:
	WERROR result;
	struct IWbemClassObject* in = NULL;
    CODE:
	result = IWbemClassObject_SpawnInstance(obj, (struct com_context*)obj, 0, &in);
	if (!in || !W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}
	RETVAL = in;
    OUTPUT:
	RETVAL

void
_Put(obj,ParamName,value)
	struct IWbemClassObject* obj;
	char *ParamName;
	SV* value;
    PREINIT:
	WERROR result;
	union CIMVAR v;
	int i;
	struct WbemClassObject* d;
	enum CIMTYPE_ENUMERATION type = 0;
    CODE:
	d = obj->object_data;
	for(i=0;i<d->obj_class->__PROPERTY_COUNT;i++)
        {
	    if(!strcmp(d->obj_class->properties[i].name,ParamName))
	    {
		type = d->obj_class->properties[i].desc->cimtype;
		break;
	    }
	}
	DEBUG_MSG("PUT: %p %s %s %i\n", obj,ParamName,SvPV_nolen(value),type);
	switch(type)
	{
	    case CIM_SINT8: v.v_sint8 = SvIV(value);break;
	    case CIM_UINT8: v.v_uint8 = SvUV(value);break;
	    case CIM_SINT16: v.v_sint16 = SvIV(value);break;
	    case CIM_UINT16: v.v_uint16 = SvUV(value);break;
	    case CIM_SINT32: v.v_sint32 = SvIV(value);break;
	    case CIM_UINT32: v.v_uint32 = SvUV(value);break;
	    case CIM_SINT64: v.v_sint64 = SvIV(value);break;
	    case CIM_UINT64: v.v_sint64 = SvUV(value);break;
	    case CIM_REAL32: v.v_real32 = SvNV(value);break;
	    case CIM_REAL64: v.v_real64 = SvNV(value);break;
	    case CIM_BOOLEAN: v.v_boolean = SvUV(value);break;
	    case CIM_STRING:
	    case CIM_DATETIME:
	    case CIM_REFERENCE: v.v_string = SvPV_nolen(value);break;
	    case CIM_CHAR16: v.v_sint16 = SvUV(value);break;
	    case CIM_OBJECT: v.v_object = NULL;break;
	}
	result = IWbemClassObject_Put(obj, (struct com_context*)obj, ParamName, 0, &v, 0);
	if (!W_ERROR_IS_OK(result)) {
	  _croak_error (wmi_errstr(result));
	}



MODULE = Net::WMI::WbemClassObject		PACKAGE = Net::WMI::WbemClassObject
PROTOTYPES: ENABLE



MODULE = Net::WMI		PACKAGE = Net::WMI
